import * as fc from '../../../../lib/fast-check';

import { func, compareFunc, compareBooleanFunc } from '../../../../src/check/arbitrary/FunctionArbitrary';
import { integer } from '../../../../src/check/arbitrary/IntegerArbitrary';

import * as genericHelper from './generic/GenericArbitraryHelper';

import * as stubRng from '../../stubs/generators';

describe('FunctionArbitrary', () => {
  describe('func', () => {
    it('Should return the same value given the same input', () =>
      fc.assert(
        fc.property(fc.integer(), fc.integer(), (seed, a) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = func(integer()).generate(mrng).value;
          return f(a) === f(a);
        })
      ));
    it('Should not depend on the ordering of calls', () =>
      fc.assert(
        fc.property(fc.integer(), fc.integer(), fc.integer(), (seed, a, b) => {
          const mrng1 = stubRng.mutable.fastincrease(seed);
          const f1 = func(integer()).generate(mrng1).value;
          const va1 = f1(a);
          const vb1 = f1(b);
          const mrng2 = stubRng.mutable.fastincrease(seed);
          const f2 = func(integer()).generate(mrng2).value;
          const vb2 = f2(b);
          const va2 = f2(a);
          return va1 === va2 && vb1 === vb2;
        })
      ));
    describe('Is valid arbitrary', () => {
      genericHelper.isValidArbitrary(() => func<[number, number], number>(integer()), {
        isEqual: (f1, f2) => f1(0, 42) === f2(0, 42),
        isValidValue: f => typeof f === 'function' && typeof f(0, 0) === 'number'
      });
    });
  });
  describe('compareFunc', () => {
    it('Should return the same value given the same input', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), (seed, a, b) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareFunc().generate(mrng).value;
          return f(a, b) === f(a, b);
        })
      ));
    it('Should be transitive', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), fc.anything(), (seed, a, b, c) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareFunc().generate(mrng).value;
          const ab = f(a, b);
          const bc = f(b, c);
          fc.pre(ab !== 0 && bc !== 0);

          if (ab < 0 && bc < 0) return f(a, c) < 0;
          else if (ab > 0 && bc > 0) return f(a, c) > 0;

          fc.pre(false);
        })
      ));
    it('Should be zero when called on a = b', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), (seed, a) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareFunc().generate(mrng).value;
          return f(a, a) === 0;
        })
      ));
    it('Should be consistent when called in reversed order', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), (seed, a, b) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareFunc().generate(mrng).value;
          const ab = f(a, b);
          const ba = f(b, a);
          if (ab === 0) return ba === 0;
          else if (ab < 0) return ba > 0;
          else return ba < 0;
        })
      ));
    describe('Is valid arbitrary', () => {
      genericHelper.isValidArbitrary(() => compareFunc(), {
        isEqual: (f1, f2) => f1({ k: 0 }, { k: 42 }) === f2({ k: 0 }, { k: 42 }),
        isValidValue: f => typeof f === 'function' && typeof f({ k: 0 }, { k: 42 }) === 'number'
      });
    });
  });
  describe('compareBooleanFunc', () => {
    it('Should return the same value given the same input', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), (seed, a, b) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareBooleanFunc().generate(mrng).value;
          return f(a, b) === f(a, b);
        })
      ));
    it('Should be transitive', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), fc.anything(), (seed, a, b, c) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareBooleanFunc().generate(mrng).value;
          const ab = f(a, b);
          const bc = f(b, c);

          if (ab && bc) return f(a, c);
          else if (!ab && !bc) return !f(a, c);

          fc.pre(false);
        })
      ));
    it('Should be false when called on a = b', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), (seed, a) => {
          const mrng = stubRng.mutable.fastincrease(seed);
          const f = compareBooleanFunc().generate(mrng).value;
          return f(a, a) === false;
        })
      ));
    it('Should be equivalent to compareFunc(a, b) < 0', () =>
      fc.assert(
        fc.property(fc.integer(), fc.anything(), fc.anything(), (seed, a, b) => {
          const mrng1 = stubRng.mutable.fastincrease(seed);
          const mrng2 = stubRng.mutable.fastincrease(seed);
          const f1 = compareBooleanFunc().generate(mrng1).value;
          const f2 = compareFunc().generate(mrng2).value;
          return f1(a, b) === f2(a, b) < 0;
        })
      ));
    describe('Is valid arbitrary', () => {
      genericHelper.isValidArbitrary(() => compareBooleanFunc(), {
        isEqual: (f1, f2) => f1({ k: 0 }, { k: 42 }) === f2({ k: 0 }, { k: 42 }),
        isValidValue: f => typeof f === 'function' && typeof f({ k: 0 }, { k: 42 }) === 'boolean'
      });
    });
  });
});
